@@ -1,51 +1,46 @@
 from unittest import TestCase
 import textwrap
 
-from jsonschema import Draft4Validator, exceptions
-from jsonschema.compat import PY3
+from jsonschema import exceptions
+from jsonschema.validators import _LATEST_VERSION
 
 
 class TestBestMatch(TestCase):
-    def best_match(self, errors):
-        errors = list(errors)
-        best = exceptions.best_match(errors)
+    def best_match_of(self, instance, schema):
+        errors = list(_LATEST_VERSION(schema).iter_errors(instance))
+        best = exceptions.best_match(iter(errors))
         reversed_best = exceptions.best_match(reversed(errors))
-        msg = "Didn't return a consistent best match!\nGot: {0}\n\nThen: {1}"
         self.assertEqual(
-            best._contents(), reversed_best._contents(),
-            msg=msg.format(best, reversed_best),
+            best._contents(),
+            reversed_best._contents(),
+            f"No consistent best match!\nGot: {best}\n\nThen: {reversed_best}",
         )
         return best
 
     def test_shallower_errors_are_better_matches(self):
-        validator = Draft4Validator(
-            {
-                "properties": {
-                    "foo": {
-                        "minProperties": 2,
-                        "properties": {"bar": {"type": "object"}},
-                    },
+        schema = {
+            "properties": {
+                "foo": {
+                    "minProperties": 2,
+                    "properties": {"bar": {"type": "object"}},
                 },
             },
-        )
-        best = self.best_match(validator.iter_errors({"foo": {"bar": []}}))
+        }
+        best = self.best_match_of(instance={"foo": {"bar": []}}, schema=schema)
         self.assertEqual(best.validator, "minProperties")
 
     def test_oneOf_and_anyOf_are_weak_matches(self):
         """
         A property you *must* match is probably better than one you have to
         match a part of.
-
         """
 
-        validator = Draft4Validator(
-            {
-                "minProperties": 2,
-                "anyOf": [{"type": "string"}, {"type": "number"}],
-                "oneOf": [{"type": "string"}, {"type": "number"}],
-            }
-        )
-        best = self.best_match(validator.iter_errors({}))
+        schema = {
+            "minProperties": 2,
+            "anyOf": [{"type": "string"}, {"type": "number"}],
+            "oneOf": [{"type": "string"}, {"type": "number"}],
+        }
+        best = self.best_match_of(instance={}, schema=schema)
         self.assertEqual(best.validator, "minProperties")
 
     def test_if_the_most_relevant_error_is_anyOf_it_is_traversed(self):
@@ -56,24 +51,51 @@ class TestBestMatch(TestCase):
 
         I.e. since only one of the schemas must match, we look for the most
         relevant one.
-
         """
 
-        validator = Draft4Validator(
-            {
-                "properties": {
-                    "foo": {
-                        "anyOf": [
-                            {"type": "string"},
-                            {"properties": {"bar": {"type": "array"}}},
-                        ],
-                    },
+        schema = {
+            "properties": {
+                "foo": {
+                    "anyOf": [
+                        {"type": "string"},
+                        {"properties": {"bar": {"type": "array"}}},
+                    ],
                 },
             },
-        )
-        best = self.best_match(validator.iter_errors({"foo": {"bar": 12}}))
+        }
+        best = self.best_match_of(instance={"foo": {"bar": 12}}, schema=schema)
         self.assertEqual(best.validator_value, "array")
 
+    def test_no_anyOf_traversal_for_equally_relevant_errors(self):
+        """
+        We don't traverse into an anyOf (as above) if all of its context errors
+        seem to be equally "wrong" against the instance.
+        """
+
+        schema = {
+            "anyOf": [
+                {"type": "string"},
+                {"type": "integer"},
+                {"type": "object"},
+            ],
+        }
+        best = self.best_match_of(instance=[], schema=schema)
+        self.assertEqual(best.validator, "anyOf")
+
+    def test_anyOf_traversal_for_single_equally_relevant_error(self):
+        """
+        We *do* traverse anyOf with a single nested error, even though it is
+        vacuously equally relevant to itself.
+        """
+
+        schema = {
+            "anyOf": [
+                {"type": "string"},
+            ],
+        }
+        best = self.best_match_of(instance=[], schema=schema)
+        self.assertEqual(best.validator, "type")
+
     def test_if_the_most_relevant_error_is_oneOf_it_is_traversed(self):
         """
         If the most relevant error is an oneOf, then we traverse its context
@@ -82,73 +104,159 @@ class TestBestMatch(TestCase):
 
         I.e. since only one of the schemas must match, we look for the most
         relevant one.
-
         """
 
-        validator = Draft4Validator(
-            {
-                "properties": {
-                    "foo": {
-                        "oneOf": [
-                            {"type": "string"},
-                            {"properties": {"bar": {"type": "array"}}},
-                        ],
-                    },
+        schema = {
+            "properties": {
+                "foo": {
+                    "oneOf": [
+                        {"type": "string"},
+                        {"properties": {"bar": {"type": "array"}}},
+                    ],
                 },
             },
-        )
-        best = self.best_match(validator.iter_errors({"foo": {"bar": 12}}))
+        }
+        best = self.best_match_of(instance={"foo": {"bar": 12}}, schema=schema)
         self.assertEqual(best.validator_value, "array")
 
+    def test_no_oneOf_traversal_for_equally_relevant_errors(self):
+        """
+        We don't traverse into an oneOf (as above) if all of its context errors
+        seem to be equally "wrong" against the instance.
+        """
+
+        schema = {
+            "oneOf": [
+                {"type": "string"},
+                {"type": "integer"},
+                {"type": "object"},
+            ],
+        }
+        best = self.best_match_of(instance=[], schema=schema)
+        self.assertEqual(best.validator, "oneOf")
+
+    def test_oneOf_traversal_for_single_equally_relevant_error(self):
+        """
+        We *do* traverse oneOf with a single nested error, even though it is
+        vacuously equally relevant to itself.
+        """
+
+        schema = {
+            "oneOf": [
+                {"type": "string"},
+            ],
+        }
+        best = self.best_match_of(instance=[], schema=schema)
+        self.assertEqual(best.validator, "type")
+
     def test_if_the_most_relevant_error_is_allOf_it_is_traversed(self):
         """
         Now, if the error is allOf, we traverse but select the *most* relevant
         error from the context, because all schemas here must match anyways.
-
         """
 
-        validator = Draft4Validator(
-            {
-                "properties": {
-                    "foo": {
-                        "allOf": [
-                            {"type": "string"},
-                            {"properties": {"bar": {"type": "array"}}},
-                        ],
-                    },
+        schema = {
+            "properties": {
+                "foo": {
+                    "allOf": [
+                        {"type": "string"},
+                        {"properties": {"bar": {"type": "array"}}},
+                    ],
                 },
             },
-        )
-        best = self.best_match(validator.iter_errors({"foo": {"bar": 12}}))
+        }
+        best = self.best_match_of(instance={"foo": {"bar": 12}}, schema=schema)
         self.assertEqual(best.validator_value, "string")
 
     def test_nested_context_for_oneOf(self):
-        validator = Draft4Validator(
-            {
-                "properties": {
-                    "foo": {
-                        "oneOf": [
-                            {"type": "string"},
-                            {
-                                "oneOf": [
-                                    {"type": "string"},
-                                    {
-                                        "properties": {
-                                            "bar": {"type": "array"},
-                                        },
+        """
+        We traverse into nested contexts (a oneOf containing an error in a
+        nested oneOf here).
+        """
+
+        schema = {
+            "properties": {
+                "foo": {
+                    "oneOf": [
+                        {"type": "string"},
+                        {
+                            "oneOf": [
+                                {"type": "string"},
+                                {
+                                    "properties": {
+                                        "bar": {"type": "array"},
                                     },
-                                ],
-                            },
-                        ],
-                    },
+                                },
+                            ],
+                        },
+                    ],
                 },
             },
-        )
-        best = self.best_match(validator.iter_errors({"foo": {"bar": 12}}))
+        }
+        best = self.best_match_of(instance={"foo": {"bar": 12}}, schema=schema)
         self.assertEqual(best.validator_value, "array")
 
+    def test_it_prioritizes_matching_types(self):
+        schema = {
+            "properties": {
+                "foo": {
+                    "anyOf": [
+                        {"type": "array", "minItems": 2},
+                        {"type": "string", "minLength": 10},
+                    ],
+                },
+            },
+        }
+        best = self.best_match_of(instance={"foo": "bar"}, schema=schema)
+        self.assertEqual(best.validator, "minLength")
+
+        reordered = {
+            "properties": {
+                "foo": {
+                    "anyOf": [
+                        {"type": "string", "minLength": 10},
+                        {"type": "array", "minItems": 2},
+                    ],
+                },
+            },
+        }
+        best = self.best_match_of(instance={"foo": "bar"}, schema=reordered)
+        self.assertEqual(best.validator, "minLength")
+
+    def test_it_prioritizes_matching_union_types(self):
+        schema = {
+            "properties": {
+                "foo": {
+                    "anyOf": [
+                        {"type": ["array", "object"], "minItems": 2},
+                        {"type": ["integer", "string"], "minLength": 10},
+                    ],
+                },
+            },
+        }
+        best = self.best_match_of(instance={"foo": "bar"}, schema=schema)
+        self.assertEqual(best.validator, "minLength")
+
+        reordered = {
+            "properties": {
+                "foo": {
+                    "anyOf": [
+                        {"type": "string", "minLength": 10},
+                        {"type": "array", "minItems": 2},
+                    ],
+                },
+            },
+        }
+        best = self.best_match_of(instance={"foo": "bar"}, schema=reordered)
+        self.assertEqual(best.validator, "minLength")
+
+    def test_boolean_schemas(self):
+        schema = {"properties": {"foo": False}}
+        best = self.best_match_of(instance={"foo": "bar"}, schema=schema)
+        self.assertIsNone(best.validator)
+
     def test_one_error(self):
-        validator = Draft4Validator({"minProperties": 2})
+        validator = _LATEST_VERSION({"minProperties": 2})
         error, = validator.iter_errors({})
         self.assertEqual(
             exceptions.best_match(validator.iter_errors({})).validator,
@@ -156,7 +264,7 @@ class TestBestMatch(TestCase):
         )
 
     def test_no_errors(self):
-        validator = Draft4Validator({})
+        validator = _LATEST_VERSION({})
         self.assertIsNone(exceptions.best_match(validator.iter_errors({})))
 
 
@@ -214,7 +322,7 @@ class TestByRelevance(TestCase):
 
 class TestErrorTree(TestCase):
     def test_it_knows_how_many_total_errors_it_contains(self):
-        # FIXME: https://github.com/Julian/jsonschema/issues/442
+        # FIXME: #442
         errors = [
             exceptions.ValidationError("Something", validator=i)
             for i in range(8)
@@ -281,7 +389,6 @@ class TestErrorTree(TestCase):
         If a validator is dumb (like :validator:`required` in draft 3) and
         refers to a path that isn't in the instance, the tree still properly
         returns a subtree for that path.
-
         """
 
         error = exceptions.ValidationError(
@@ -290,22 +397,36 @@ class TestErrorTree(TestCase):
         tree = exceptions.ErrorTree([error])
         self.assertIsInstance(tree["foo"], exceptions.ErrorTree)
 
+    def test_repr(self):
+        e1, e2 = (
+            exceptions.ValidationError(
+                "1",
+                validator="foo",
+                path=["bar", "bar2"],
+                instance="i1"),
+            exceptions.ValidationError(
+                "2",
+                validator="quux",
+                path=["foobar", 2],
+                instance="i2"),
+        )
+        tree = exceptions.ErrorTree([e1, e2])
+        self.assertEqual(repr(tree), "<ErrorTree (2 total errors)>")
+
 
 class TestErrorInitReprStr(TestCase):
     def make_error(self, **kwargs):
         defaults = dict(
-            message=u"hello",
-            validator=u"type",
-            validator_value=u"string",
+            message="hello",
+            validator="type",
+            validator_value="string",
             instance=5,
-            schema={u"type": u"string"},
+            schema={"type": "string"},
         )
         defaults.update(kwargs)
         return exceptions.ValidationError(**defaults)
 
     def assertShows(self, expected, **kwargs):
-        if PY3:  # pragma: no cover
-            expected = expected.replace("u'", "'")
         expected = textwrap.dedent(expected).rstrip("\n")
 
         error = self.make_error(**kwargs)
@@ -320,7 +441,7 @@ class TestErrorInitReprStr(TestCase):
     def test_repr(self):
         self.assertEqual(
             repr(exceptions.ValidationError(message="Hello!")),
-            "<ValidationError: %r>" % "Hello!",
+            "<ValidationError: 'Hello!'>",
         )
 
     def test_unset_error(self):
@@ -343,8 +464,8 @@ class TestErrorInitReprStr(TestCase):
     def test_empty_paths(self):
         self.assertShows(
             """
-            Failed validating u'type' in schema:
-                {u'type': u'string'}
+            Failed validating 'type' in schema:
+                {'type': 'string'}
 
             On instance:
                 5
@@ -356,8 +477,8 @@ class TestErrorInitReprStr(TestCase):
     def test_one_item_paths(self):
         self.assertShows(
             """
-            Failed validating u'type' in schema:
-                {u'type': u'string'}
+            Failed validating 'type' in schema:
+                {'type': 'string'}
 
             On instance[0]:
                 5
@@ -369,20 +490,20 @@ class TestErrorInitReprStr(TestCase):
     def test_multiple_item_paths(self):
         self.assertShows(
             """
-            Failed validating u'type' in schema[u'items'][0]:
-                {u'type': u'string'}
+            Failed validating 'type' in schema['items'][0]:
+                {'type': 'string'}
 
-            On instance[0][u'a']:
+            On instance[0]['a']:
                 5
             """,
-            path=[0, u"a"],
-            schema_path=[u"items", 0, 1],
+            path=[0, "a"],
+            schema_path=["items", 0, 1],
         )
 
     def test_uses_pprint(self):
         self.assertShows(
             """
-            Failed validating u'maxLength' in schema:
+            Failed validating 'maxLength' in schema:
                 {0: 0,
                  1: 1,
                  2: 2,
@@ -433,15 +554,14 @@ class TestErrorInitReprStr(TestCase):
             """,
             instance=list(range(25)),
             schema=dict(zip(range(20), range(20))),
-            validator=u"maxLength",
+            validator="maxLength",
         )
 
     def test_str_works_with_instances_having_overriden_eq_operator(self):
         """
-        Check for https://github.com/Julian/jsonschema/issues/164 which
-        rendered exceptions unusable when a `ValidationError` involved
-        instances with an `__eq__` method that returned truthy values.
-
+        Check for #164 which rendered exceptions unusable when a
+        `ValidationError` involved instances with an `__eq__` method
+        that returned truthy values.
         """
 
         class DontEQMeBro(object):
